/*
  CPP_CONTEST=2013
  CPP_PROBLEM=E
  CPP_LANG=CUDA
  CPP_PROCESSES_PER_NODE=mercurio 1
*/

/* RECORD
Juan Mompeán Esteban
students of the University of Murcia, Spain
In the warm-up for the IV Spanish Parallel Programming Contest
time 3340 msec
speed-up 4.04
*/ 

/*The first CUDA solution for the problem.
*/

#include <stdlib.h>
#include <stdio.h>
#include <helper_functions.h>
#include <helper_cuda.h>

__global__ void reduce0(double *g_idata, double *g_odata, int N) {
  extern __shared__ double sdata[];
  // each thread loads one element from global to shared mem
  unsigned int tid = threadIdx.x;
  unsigned int i = blockIdx.x * blockDim.x + threadIdx.x;
  unsigned int j = blockIdx.y * blockDim.y + threadIdx.y;
  if (i >= N || j >= N) {
    return;
  }
  sdata[tid] = g_idata[j * N + i];
  __syncthreads();
  if (i + j > N) {
    return;
  }
  // do reduction in shared mem
  for (unsigned int s = blockDim.x / 2; s > 0; s >>= 1) {
    if (tid < s) {
      sdata[tid] += sdata[tid + s];
    } else {
      break;
    }
    __syncthreads();
    if (tid == 0 && s % 2 == 1 && s != 1) {
      sdata[0] += sdata[s - 1];
    }
  }
  // write result for this block to global mem
  if (tid == 0) {
    g_odata[j * N + blockIdx.x] = sdata[0];
  }
}

__global__ void reduce1(double *g_idata, double *g_odata, int N) {
  extern __shared__ double sdata[];
  // each thread loads one element from global to shared mem
  unsigned int tid = threadIdx.x;
  unsigned int i = blockIdx.x * blockDim.x + threadIdx.x;
  if (i >= N) {
    return;
  }
  sdata[tid] = g_idata[i * N];
  __syncthreads();
  // do reduction in shared mem
  for (unsigned int s = blockDim.x / 2; s > 0; s >>= 1) {
    if (tid < s) {
      sdata[tid] += sdata[tid + s];
    }
    __syncthreads();
    if (tid == 0 && s % 2 == 1 && s != 1) {
      sdata[tid] += sdata[s - 1];
    }
  }
  // write result for this block to global mem
  if (tid == 0) {
    g_odata[i * N] = sdata[0];
  }
}

__global__ void oneTaylor1(int N, double *v, double *a, double x, double y) {
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  int j = blockIdx.y * blockDim.y + threadIdx.y;
  if (i >= N || j >= N || i + j >= N) {
    return;
  }
  int idx = i * N + j;
  v[idx] = a[idx] * (1 / (pow(y, j) * pow(x, i)));
}
 
void  sec(int N,double *a,int P,double *x,double *y,double *r)
{
  int i;
  double *d_v, *d_a, *v;
  v = (double*)malloc(sizeof(double) * N * N);
  cudaMalloc(&d_v, sizeof(double) * N * N);
  cudaMalloc(&d_a, sizeof(double) * N * N);
  cudaMemset(d_v, 0, sizeof(double) * N * N);
  cudaMemcpy(d_a, a, sizeof(double) * N * N, cudaMemcpyHostToDevice);
  for(i=0;i<P*P;i++)
  {
    dim3 block(8, 8, 1);
    dim3 grid(ceil(N / (float)block.x), ceil (N / (float)block.y), 1);
    oneTaylor1<<<grid, block>>>(N, d_v, d_a, x[i], y[i]);
    int N2 = N;
    
    do {
      block.x = min(512, N2);
      block.y = 1;
      grid.x = ceil(N2 / (float)block.x);
      grid.y = N;
      reduce0<<<grid, block, block.x * sizeof(double)>>>(d_v, d_v, N2);
      N2 = grid.x;
    } while(grid.x > 1);
    N2 = N;
    do {
      block.x = min(512, N2);
      block.y = 1;
      grid.x = ceil(N2 / (float)block.x);
      grid.y = 1;
      reduce1<<<grid, block, block.x * sizeof(double)>>>(d_v, d_v, N2);
      N2 = grid.x;
    } while(grid.x > 1);

    cudaMemcpy(v, d_v, sizeof(double), cudaMemcpyDeviceToHost);
    r[i] = v[0];
  }
  free(v);
  cudaFree(d_v);
}
